##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import base64
import email
import unittest

import six

from DocumentTemplate.DT_Util import ParseError


class MimeTest(unittest.TestCase):

    def _getTargetClass(self):
        from Products.MIMETools import MIMETag
        return MIMETag.MIMETag

    def _makeOne(self, blocks=[]):
        klass = self._getTargetClass()
        return klass(blocks)

    def _decode64(self, text):
        if six.PY3 and isinstance(text, str):
            text = text.encode('utf-8')
        try:
            return base64.decodebytes(text)
        except AttributeError:
            return base64.decodestring(text)

    @property
    def doc_class(self):
        from DocumentTemplate.DT_HTML import HTML
        return HTML

    def test_registered(self):
        klass = self._getTargetClass()
        from DocumentTemplate.DT_String import String
        self.assertTrue('mime' in String.commands)
        self.assertTrue(String.commands['mime'] is klass)

    def test_init(self):
        tag = self._makeOne()
        self.assertEqual(tag.sections, [])

    def test_render(self):
        tag = self._makeOne()
        result = tag.render(md={})
        self.assertIn('MIME-Version: 1.0', result)
        self.assertIn('Content-Type: multipart/mixed;', result)

    def test_call(self):
        tag = self._makeOne()
        result = tag(md={})
        self.assertIn('MIME-Version: 1.0', result)
        self.assertIn('Content-Type: multipart/mixed;', result)

    def test_text(self):
        html = self.doc_class(
            '<dtml-mime type="text/html" charset="utf-8">'
            '<b>I am BOLD</b>'
            '<dtml-boundary type="text/plain">'
            'Just plain text'
            '<dtml-boundary type="text/css" disposition="attachment" '
            '               filename="my.css" name="my" charset="iso8859-1" '
            '               cid="CID">'
            'All bells and whistles'
            '<dtml-boundary skip_expr="1==1">'
            'You cannot see me'
            '</dtml-mime>')

        msg = email.message_from_string(html(md={}))
        self.assertTrue(msg.is_multipart())

        parts = msg.get_payload()
        self.assertEqual(len(parts), 3)

        part1 = parts[0]
        self.assertFalse(part1.is_multipart())
        self.assertEqual(part1['Content-Type'],
                         'text/html; charset="utf-8"')
        self.assertEqual(part1['MIME-Version'], '1.0')
        self.assertEqual(part1['Content-Transfer-Encoding'], 'base64')
        self.assertEqual(self._decode64(part1.get_payload()),
                         b'<b>I am BOLD</b>')

        part2 = parts[1]
        self.assertFalse(part2.is_multipart())
        self.assertEqual(part2['Content-Type'],
                         'text/plain; charset="us-ascii"')
        self.assertEqual(part2['MIME-Version'], '1.0')
        self.assertEqual(part2['Content-Transfer-Encoding'], '7bit')
        self.assertEqual(part2.get_payload(), 'Just plain text')

        part3 = parts[2]
        self.assertFalse(part3.is_multipart())
        self.assertEqual(part3['Content-Type'],
                         'text/css; charset="iso8859-1"')
        self.assertEqual(part3['MIME-Version'], '1.0')
        self.assertEqual(part3['Content-Transfer-Encoding'], 'base64')
        self.assertEqual(part3['Content-ID'], '<CID>')
        self.assertEqual(part3['Content-Disposition'],
                         'attachment;\n filename="my.css"')
        self.assertEqual(self._decode64(part3.get_payload()),
                         b'All bells and whistles')

    def test_other(self):
        html = self.doc_class(
            '<dtml-mime>'
            '<b>I am BOLD</b>'
            '<dtml-boundary type="dont/know">'
            'Just plain text'
            '<dtml-boundary type="application/js" disposition="attachment" '
            '               filename="my.js" name="my" '
            '               encode="quoted-printable" cid="CID">'
            'All bells and whistles'
            '</dtml-mime>')

        msg = email.message_from_string(html(md={}))
        self.assertTrue(msg.is_multipart())

        parts = msg.get_payload()
        self.assertEqual(len(parts), 3)

        part1 = parts[0]
        self.assertFalse(part1.is_multipart())
        self.assertEqual(part1['Content-Type'],
                         'application/octet-stream')
        self.assertEqual(part1['MIME-Version'], '1.0')
        self.assertEqual(part1['Content-Transfer-Encoding'], 'base64')
        self.assertEqual(self._decode64(part1.get_payload()),
                         b'<b>I am BOLD</b>')

        part2 = parts[1]
        self.assertFalse(part2.is_multipart())
        self.assertEqual(part2['Content-Type'], 'application/octet-stream')
        self.assertEqual(part2['MIME-Version'], '1.0')
        self.assertEqual(part2['Content-Transfer-Encoding'], 'base64')
        self.assertEqual(self._decode64(part2.get_payload()),
                         b'Just plain text')

        part3 = parts[2]
        self.assertFalse(part3.is_multipart())
        self.assertEqual(part3['Content-Type'], 'application/js')
        self.assertEqual(part3['MIME-Version'], '1.0')
        self.assertEqual(part3['Content-Transfer-Encoding'],
                         'quoted-printable')
        self.assertEqual(part3['Content-ID'], '<CID>')
        self.assertEqual(part3['Content-Disposition'],
                         'attachment;\n filename="my.js"')
        self.assertEqual(part3.get_payload(),
                         'All=20bells=20and=20whistles')

    def test_bad_encoding(self):
        from .MIMETag import MIMEError
        broken = '<dtml-mime type="image/jpg" encode="foo"></dtml-mime>'
        html = self.doc_class(broken)
        with self.assertRaises(MIMEError) as cm:
            html()
        self.assertIn('unsupported encoding', str(cm.exception))

    def test_forbidden_combined_attributes(self):
        broken = '<dtml-mime type="x" type_expr="y"></dtml-mime>'
        html = self.doc_class(broken)
        with self.assertRaises(ParseError) as cm:
            html()
        self.assertIn('type and type_expr given', str(cm.exception))

        broken = ('<dtml-mime disposition="x" '
                  'disposition_expr="y"></dtml-mime>')
        html = self.doc_class(broken)
        with self.assertRaises(ParseError) as cm:
            html()
        self.assertIn('disposition and disposition_expr given',
                      str(cm.exception))

        broken = '<dtml-mime encode="x" encode_expr="y"></dtml-mime>'
        html = self.doc_class(broken)
        with self.assertRaises(ParseError) as cm:
            html()
        self.assertIn('encode and encode_expr given', str(cm.exception))

        broken = '<dtml-mime name="x" name_expr="y"></dtml-mime>'
        html = self.doc_class(broken)
        with self.assertRaises(ParseError) as cm:
            html()
        self.assertIn('name and name_expr given', str(cm.exception))

        broken = '<dtml-mime filename="x" filename_expr="y"></dtml-mime>'
        html = self.doc_class(broken)
        with self.assertRaises(ParseError) as cm:
            html()
        self.assertIn('filename and filename_expr given', str(cm.exception))

        broken = '<dtml-mime cid="x" cid_expr="y"></dtml-mime>'
        html = self.doc_class(broken)
        with self.assertRaises(ParseError) as cm:
            html()
        self.assertIn('cid and cid_expr given', str(cm.exception))

        broken = '<dtml-mime charset="x" charset_expr="y"></dtml-mime>'
        html = self.doc_class(broken)
        with self.assertRaises(ParseError) as cm:
            html()
        self.assertIn('charset and charset_expr given', str(cm.exception))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MimeTest))
    return suite



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from email.encoders import encode_7or8bit
from email.encoders import encode_base64
from email.encoders import encode_quopri
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from io import BytesIO
from io import StringIO

import six

from DocumentTemplate._DocumentTemplate import render_blocks
from DocumentTemplate.DT_String import String
from DocumentTemplate.DT_Util import Eval
from DocumentTemplate.DT_Util import ParseError
from DocumentTemplate.DT_Util import parse_params


if six.PY2:
    outfile = BytesIO
else:
    outfile = StringIO
TYPE_CLASSES = {
    'text': MIMEText,
    'image': MIMEImage,
    'audio': MIMEAudio,
    'application': MIMEApplication,
}
ENCODINGS = {
    '7bit': encode_7or8bit,
    '8bit': encode_7or8bit,
    'base64': encode_base64,
    'quoted-printable': encode_quopri,
}


class MIMEError(Exception):
    """MIME Tag Error"""


class MIMETag(object):
    """ The dtml-mime tag """

    name = 'mime'
    blockContinuations = ('boundary',)
    encode = None

    def __init__(self, blocks, encoding=None):
        self.encoding = encoding
        self.sections = []
        self.multipart = 'mixed'

        for tname, args, section in blocks:
            if tname == 'mime':
                args = parse_params(args, type=None, type_expr=None,
                                    disposition=None, disposition_expr=None,
                                    encode=None, encode_expr=None, name=None,
                                    name_expr=None, filename=None,
                                    filename_expr=None, cid=None,
                                    cid_expr=None, charset=None,
                                    charset_expr=None, skip_expr=None,
                                    multipart=None)
                self.multipart = args.get('multipart', 'mixed')
            else:
                args = parse_params(args, type=None, type_expr=None,
                                    disposition=None, disposition_expr=None,
                                    encode=None, encode_expr=None, name=None,
                                    name_expr=None, filename=None,
                                    filename_expr=None, cid=None,
                                    cid_expr=None, charset=None,
                                    charset_expr=None, skip_expr=None)

            if 'type_expr' in args:
                if 'type' in args:
                    raise ParseError('dtml-mime: type and type_expr given')
                args['type_expr'] = Eval(args['type_expr'])
            elif 'type' not in args:
                args['type'] = 'application/octet-stream'

            if 'disposition_expr' in args:
                if 'disposition' in args:
                    raise ParseError(
                        'dtml-mime: disposition and disposition_expr given')
                args['disposition_expr'] = Eval(args['disposition_expr'])
            elif 'disposition' not in args:
                args['disposition'] = ''

            if 'encode_expr' in args:
                if 'encode' in args:
                    raise ParseError('dtml-mime: encode and encode_expr given')
                args['encode_expr'] = Eval(args['encode_expr'])
            elif 'encode' not in args:
                args['encode'] = 'base64'

            if 'name_expr' in args:
                if 'name' in args:
                    raise ParseError('dtml-mime: name and name_expr given')
                args['name_expr'] = Eval(args['name_expr'])
            elif 'name' not in args:
                args['name'] = ''

            if 'filename_expr' in args:
                if 'filename' in args:
                    raise ParseError(
                        'dtml-mime: filename and filename_expr given')
                args['filename_expr'] = Eval(args['filename_expr'])
            elif 'filename' not in args:
                args['filename'] = ''

            if 'cid_expr' in args:
                if 'cid' in args:
                    raise ParseError('dtml-mime: cid and cid_expr given')
                args['cid_expr'] = Eval(args['cid_expr'])
            elif 'cid' not in args:
                args['cid'] = ''

            if 'charset_expr' in args:
                if 'charset' in args:
                    raise ParseError(
                        'dtml-mime: charset and charset_expr given')
                args['charset_expr'] = Eval(args['charset_expr'])
            elif 'charset' not in args:
                args['charset'] = 'us-ascii'  # Default for text parts

            if 'skip_expr' in args:
                args['skip_expr'] = Eval(args['skip_expr'])

            if args['encode'] not in ENCODINGS:
                raise MIMEError('An unsupported encoding was specified in tag')

            self.sections.append((args, section.blocks))

    def render(self, md):
        outer = MIMEMultipart(self.multipart)

        for (args, blocks) in self.sections:
            if 'skip_expr' in args and args['skip_expr'].eval(md):
                continue

            if 'type_expr' in args:
                typ = args['type_expr'].eval(md)
            else:
                typ = args['type']

            if 'disposition_expr' in args:
                disposition = args['disposition_expr'].eval(md)
            else:
                disposition = args['disposition']

            if 'encode_expr' in args:
                encode = args['encode_expr'].eval(md)
            else:
                encode = args['encode']

            if 'filename_expr' in args:
                filename = args['filename_expr'].eval(md)
            else:
                filename = args['filename']

            if 'cid_expr' in args:
                cid = args['cid_expr'].eval(md)
            else:
                cid = args['cid']

            if 'charset_expr' in args:
                charset = args['charset_expr'].eval(md)
            else:
                charset = args['charset']

            maintype, subtype = [x.lower() for x in typ.split('/')]
            if maintype not in TYPE_CLASSES:
                maintype = 'application'
                subtype = 'octet-stream'

            klass = TYPE_CLASSES.get(maintype, MIMEApplication)
            data = render_blocks(blocks, md)

            if maintype == 'text':
                inner = klass(data, _subtype=subtype, _charset=charset)
            else:
                inner = klass(data, _subtype=subtype,
                              _encoder=ENCODINGS.get(encode))

            if cid:
                inner.add_header('Content-ID', '<%s>' % cid)

            if disposition:
                if filename:
                    inner.add_header('Content-Disposition',
                                     '%s;\n filename="%s"' % (disposition,
                                                              filename))
                else:
                    inner.add_header('Content-Disposition', disposition)

            outer.attach(inner)

        return outer.as_string()

    __call__ = render


String.commands['mime'] = MIMETag



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from setuptools import find_packages
from setuptools import setup


setup(name='Products.MIMETools',
      version='3.2',
      url='https://github.com/zopefoundation/Products.MIMETools',
      project_urls={
        'Issue Tracker': ('https://github.com/zopefoundation'
                          '/Products.MIMETools/issues'),
        'Sources': 'https://github.com/zopefoundation/Products.MIMETools',
      },
      license='ZPL 2.1',
      description='MIMETools provides the ``dtml-mime`` tag for '
                  'DocumentTemplate.',
      author='Zope Foundation and Contributors',
      author_email='zope-dev@zope.org',
      long_description=(open('README.rst').read() + '\n' +
                        open('CHANGES.txt').read()),
      packages=find_packages('src'),
      namespace_packages=['Products'],
      package_dir={'': 'src'},
      classifiers=[
        'Development Status :: 6 - Mature',
        'Environment :: Web Environment',
        'Framework :: Zope',
        'Framework :: Zope :: 4',
        'Framework :: Zope :: 5',
        'License :: OSI Approved :: Zope Public License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Communications :: Email',
      ],
      python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
      install_requires=[
        'setuptools',
        'six',
        'ExtensionClass>=4.1a1',
        'DocumentTemplate>=3',
      ],
      include_package_data=True,
      zip_safe=False,
      )
